/* SPDX-License-Identifier: LGPL-2.1-or-later */
/*
 * Copyright (C) 2023 Javier Sánchez Parra, javsanpar@riseup.net
 */

#ifndef NMTUI_RADIO_H
#define NMTUI_RADIO_H

NmtNewtForm *nmtui_radio(gboolean is_top, int argc, char **argv);

#endif /* NMTUI_RADIO_H */
