/* SPDX-License-Identifier: GPL-2.0-or-later */
/*
 * Copyright (C) 2011 Red Hat, Inc.
 */

#ifndef __NETWORKMANAGER_DEVICE_ETHERNET_UTILS_H__
#define __NETWORKMANAGER_DEVICE_ETHERNET_UTILS_H__

char *nm_device_ethernet_utils_get_default_wired_name(GHashTable *existing_ids);

#endif /* NETWORKMANAGER_DEVICE_ETHERNET_UTILS_H */
